package ch.vanwa.lox_interpreter;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;

/**
 * A Lox interpreter.
 */
public class Lox {
  /**
   * True if a run produced an error, false otherwise.
   */
  static boolean hadError = false;

  /**
   * Run a file with Lox source code.
   *
   * @param path Path of the source code file.
   */
  public static void runFile(String path) throws IOException {
    byte[] bytes = Files.readAllBytes(Paths.get(path));
    run(new String(bytes, StandardCharsets.UTF_8));
    if (hadError) {
      System.exit(65);
    }
  }

  /**
   * Run a Lox REPL.
   */
  public static void runPrompt() throws IOException {
    var input = new InputStreamReader(System.in);
    var reader = new BufferedReader(input);

    for (;;) {
      System.out.print("> ");
      String line = reader.readLine();
      if (line == null) {
        break;
      }
      run(line);
      hadError = false;
    }
  }

  /**
   * Run Lox source code.
   *
   * @param source Lox source code.
   *
   * @return List of scanned tokens.
   */
  public static void run(String source) {
    var scanner = new Scanner(source);
    List<Token> tokens = scanner.scanTokens();
    Parser parser = new Parser(tokens);
    Expr expression = parser.parse();
    if (hadError)
      return;

    System.out.println(new AstPrinter().print(expression));
  }

  /**
   * Print a simple error message.
   *
   * @param line    Line number where the error ocurred.
   * @param message Error message.
   */
  static void error(int line, String message) {
    report(line, "", message);
  }

  /**
   * Print an error message for a token.
   *
   * @param line    Line number where the error ocurred.
   * @param message Error message.
   */
  static void error(Token token, String message) {
    if (token.type() == TokenType.EOF) {
      report(token.line(), " at end", message);
    } else {
      report(token.line(), " at '" + token.lexeme() + "'", message);
    }
  }

  /**
   * Print a message to stderr.
   *
   * @param line    Line number where the error ocurred.
   * @param where   Where the error occurred.
   * @param message Error message.
   */
  private static void report(int line, String where, String message) {
    System.err.println(String.format("[line %d] Error%s: %s", line, where, message));
    hadError = true;
  }
}
